from functools import partial
from math import ceil, floor
from typing import Callable, Final, Literal, cast

from riot_na.data.model import AlignmentString, ChainType, Scheme
from riot_na.data.scheme_definitions import get_indels_positions
from riot_na.data.scheme_regions import get_cdr_ranges
from riot_na.schemes.collapse_alignment import collapse_ins_del

Op = Literal["M", "D", "N", "I", "S"]
M: Final[Op] = "M"
D: Final[Op] = "D"
N: Final[Op] = "N"
I: Final[Op] = "I"
S: Final[Op] = "S"


def _reorder_cdr_deletions(relative_indel_position: int, matches: int, remaining_deletions: int) -> str:
    # deletions should be put to the left of specified position
    # indel_position = matches_before + deletions => matches_before = indel_position - deletions
    matches_before = relative_indel_position - remaining_deletions
    if matches_before >= 0:
        assert matches_before <= matches
        return f"{matches_before * M}{remaining_deletions * D}{(matches - matches_before) * M}"

    # if there are too many deletions, just put deletions first, then matches
    return f"{remaining_deletions * D}{matches  * M}"


def _reorder_cdr_deletions_imgt(relative_indel_position: int, matches: int, remaining_deletions: int) -> str:
    # deletions should be put in the middle of cdr, starting from indel_position + 1
    # for cdr1 and cdr2, and indel_position for cdr3. We can detect cdr number from
    # alignment string length, cdr1 and cdr2 have even number of positions, while cdr3 odd.
    legal_positions = matches + remaining_deletions
    round_fn = ceil if legal_positions % 2 else floor
    matches_before = relative_indel_position - round_fn(remaining_deletions / 2)
    matches_before = max(matches_before, 0)
    assert matches >= matches_before
    return f"{matches_before * M}{remaining_deletions * D}{(matches - matches_before) * M}"


def _reorder_cdr(
    alignment_str: AlignmentString,
    relative_indel_position: int,
    reorder_buffer_deletions: Callable[[int, int, int], str],
) -> AlignmentString:
    matches = alignment_str.count(M)
    dels = alignment_str.count(D)
    ins = alignment_str.count(I)

    assert matches + dels + ins == len(alignment_str)

    extra_matches = min(ins, dels)
    matches = matches + extra_matches

    # collapse insertions and deletions over the alignment
    if ins > dels:
        remaining_insertions = ins - dels
        # insertions should be put after specified position or in the middle
        assert relative_indel_position <= matches
        result = f"{relative_indel_position * M}{remaining_insertions * I}{(matches - relative_indel_position) * M}"

    elif dels > ins:
        remaining_deletions = dels - ins
        result = reorder_buffer_deletions(relative_indel_position, matches, remaining_deletions)
    else:
        result = matches * M

    return AlignmentString(result)


reorder_buffer_other_schemes = cast(
    Callable[[AlignmentString, int], AlignmentString],
    partial(_reorder_cdr, reorder_buffer_deletions=_reorder_cdr_deletions),
)
reorder_buffer_imgt = cast(
    Callable[[AlignmentString, int], AlignmentString],
    partial(_reorder_cdr, reorder_buffer_deletions=_reorder_cdr_deletions_imgt),
)


def smooth_cdr_junctions(
    alignment_str: AlignmentString, chain_type: ChainType, scheme: Scheme = Scheme.IMGT
) -> AlignmentString:
    """
    Function for reordering function for reordering cdr insertions and deletions
    to match numbering scheme standard.

    Each numbering scheme (IMGT, KABAT, CHOTHIA, MARTIN) specifies legal positions
    in CDR regions, where insertions or deletions should be put. This function takes
    each CDR and reorders alignment string to match those specifications.

    IMGT standard: https://www.imgt.org/IMGTScientificChart/Numbering/IMGTIGVLsuperfamily.html
    KABAT, CHOTHIA, MARTIN: http://www.bioinf.org.uk/abs/info.html
    """

    reorder_buffer_fn: Callable[[AlignmentString, int], AlignmentString] = (
        reorder_buffer_imgt if scheme == Scheme.IMGT else reorder_buffer_other_schemes
    )
    scheme_position = 0
    smoothed = ""
    buffer = ""
    regions_it = iter(get_cdr_ranges(scheme, chain_type))
    indel_positions_it = iter(get_indels_positions(scheme, chain_type))

    region_start, region_end = next(regions_it, (None, None))
    indel_position = next(indel_positions_it, None)

    for op in alignment_str:
        if op in (M, D, N):
            scheme_position = scheme_position + 1

        # Append the rest of fwr4
        if not (region_start and region_end and indel_position):
            smoothed = smoothed + op
            continue

        # Take only cdr + insertions right before cdr start
        if (region_start - 1 == scheme_position and op == I) or region_start <= scheme_position <= region_end:
            buffer = buffer + op
        else:
            if buffer:
                # For cases when alignment starts on framework, do nothing.
                if set(buffer) == {N}:
                    smoothed = smoothed + buffer
                else:
                    # +1 because of 1-based indexing
                    relative_indel_position = indel_position - region_start + 1
                    fixed_buffer = reorder_buffer_fn(
                        AlignmentString(buffer.replace(N, D)),
                        relative_indel_position,
                    )
                    smoothed = smoothed + fixed_buffer

                buffer = ""
                region_start, region_end = next(regions_it, (None, None))
                indel_position = next(indel_positions_it, None)

            smoothed = smoothed + op

    # For cases when alignment ends on cdr
    if buffer:
        assert N not in buffer, "No framework aligned"
        assert region_start and region_end and indel_position

        region_length = region_end - region_start + 1
        collapsed_buffer = collapse_ins_del(buffer)
        missing_deletions_count = region_length - (collapsed_buffer.count(M) + collapsed_buffer.count(D))
        filled_buffer = AlignmentString(f"{collapsed_buffer}{missing_deletions_count * D}")
        relative_indel_position = indel_position - region_start + 1
        fixed_buffer = reorder_buffer_fn(filled_buffer, relative_indel_position)
        smoothed = smoothed + fixed_buffer

    return AlignmentString(smoothed)



from typing import Literal

import pandas as pd


def calculate_gene_allele_assignment_precision(
    ground_truth_df: pd.DataFrame, df: pd.DataFrame, gene: Literal["v", "j", "d", "c"] = "v"
) -> dict[str, float]:
    joined = ground_truth_df[[f"{gene}_call"]].join(df[[f"{gene}_call"]], lsuffix="_true", how="inner")
    return {
        "gene": (joined[f"{gene}_call_true"].str.split("*").str[0] == joined[f"{gene}_call"].str.split("*").str[0])
        .value_counts(normalize=True)
        .loc[True]
        * 100,
        "allele": (joined[f"{gene}_call_true"] == joined[f"{gene}_call"]).value_counts(normalize=True).loc[True] * 100,
    }



from riot_na.data.model import AlignmentString, ChainType, Scheme
from riot_na.schemes.smooth_alignment import (
    _reorder_cdr_deletions_imgt,
    smooth_cdr_junctions,
)

SCHEME = Scheme.IMGT
CHAIN_TYPE = ChainType.HEAVY


def test_smooth_cdr_junctions_imgt_no_indels():
    regions = {
        "fwr1": "MMMMMMMMMMMMMMMMMMMMMMMMMM",
        "cdr1": "MMMMMMMMMMMM",
        "fwr2": "MMMMMMMMMMMMMMMMM",
        "cdr2": "MMMMMMMMMM",
        "fwr3": "MMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMM",
        "cdr3": "MMMMMMMMMMMMM",
        "fwr4": "MMMMMMMMMMM",
    }
    given_alignment_str = AlignmentString("".join(regions.values()))
    assert sum(map(given_alignment_str.count, ["M", "D", "N"])) == 128

    assert (
        smooth_cdr_junctions(alignment_str=given_alignment_str, chain_type=CHAIN_TYPE, scheme=SCHEME)
        == given_alignment_str
    )


def test_smooth_cdr_junctions_deletions_not_in_cdr():
    regions = {
        "fwr1": "MMMMMMMMMMMMMMMMMMMMMMMDMD",
        "cdr1": "MMMMMMMMMMMM",
        "fwr2": "DMMMMMMMMMMDMMMMD",
        "cdr2": "MMMMMMMMMM",
        "fwr3": "MMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMDD",
        "cdr3": "MMMMMMMMMMMMM",
        "fwr4": "DMMMMDMMMMM",
    }
    given_alignment_str = AlignmentString("".join(regions.values()))
    assert sum(map(given_alignment_str.count, ["M", "D", "N"])) == 128

    regions = {
        "fwr1": "MMMMMMMMMMMMMMMMMMMMMMMDMD",
        "cdr1": "MMMMMMMMMMMM",
        "fwr2": "DMMMMMMMMMMDMMMMD",
        "cdr2": "MMMMMMMMMM",
        "fwr3": "MMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMDD",
        "cdr3": "MMMMMMMMMMMMM",
        "fwr4": "DMMMMDMMMMM",
    }
    then_alignment_str = AlignmentString("".join(regions.values()))
    assert sum(map(then_alignment_str.count, ["M", "D", "N"])) == 128

    assert (
        smooth_cdr_junctions(alignment_str=given_alignment_str, chain_type=CHAIN_TYPE, scheme=SCHEME)
        == then_alignment_str
    )


def test_smooth_cdr_junctions_imgt_deletions_in_cdrs():
    regions = {
        "fwr1": "MMMMMMMMMMMMMMMMMMMMMMMDMD",
        "cdr1": "MMDMMMMMMDMM",
        "fwr2": "DMMMMMMMMMMDMMMMD",
        "cdr2": "MMMMMMMMMD",
        "fwr3": "MMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMDD",
        "cdr3": "MMMMMMMMMMDMM",
        "fwr4": "DMMMMDMMMMM",
    }
    given_alignment_str = AlignmentString("".join(regions.values()))
    assert sum(map(given_alignment_str.count, ["M", "D", "N"])) == 128

    regions = {
        "fwr1": "MMMMMMMMMMMMMMMMMMMMMMMDMD",
        "cdr1": "MMMMMDDMMMMM",
        "fwr2": "DMMMMMMMMMMDMMMMD",
        "cdr2": "MMMMMDMMMM",
        "fwr3": "MMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMDD",
        "cdr3": "MMMMMMDMMMMMM",
        "fwr4": "DMMMMDMMMMM",
    }
    then_alignment_str = AlignmentString("".join(regions.values()))
    assert sum(map(then_alignment_str.count, ["M", "D", "N"])) == 128

    assert (
        smooth_cdr_junctions(alignment_str=given_alignment_str, chain_type=CHAIN_TYPE, scheme=SCHEME)
        == then_alignment_str
    )


def test_smooth_cdr_junctions_imgt_insertions_in_frameworks():
    regions = {
        "fwr1": "MIIMMMMMMMMMMMMMMIMMMMMMMMMIMIM",
        "cdr1": "MMMMMMMMMMMM",
        "fwr2": "MIIIMMMMMMIMMMMIMMMMMIIM",
        "cdr2": "MMMMMMMMMM",
        "fwr3": "MIIMMMMMMMMMMMMMIIMMMMMMMMMMMMMIIMMMMMMMMMMMIIM",
        "cdr3": "MMMMMMMMMMMMM",
        "fwr4": "MMMMMMMMMMM",
    }
    given_alignment_str = AlignmentString("".join(regions.values()))
    assert sum(map(given_alignment_str.count, ["M", "D", "N"])) == 128

    assert (
        smooth_cdr_junctions(alignment_str=given_alignment_str, chain_type=CHAIN_TYPE, scheme=SCHEME)
        == given_alignment_str
    )


def test_smooth_cdr_junctions_imgt_insertions_in_cdrs():
    regions = {
        "fwr1": "MMMMMMMMMMMMMMMMMMMMMMMMMMII",
        "cdr1": "MIMMMMMMMMIMMM",
        "fwr2": "IIMMMMMMMMMMMMMMMMM",
        "cdr2": "MMIMMMMMMMMI",
        "fwr3": "IMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMM",
        "cdr3": "IMMIMMMMMMMMMIIMM",
        "fwr4": "IIMMMMMMMMMMM",
    }
    given_alignment_str = AlignmentString("".join(regions.values()))
    assert sum(map(given_alignment_str.count, ["M", "D", "N"])) == 128

    regions = {
        "fwr1": "MMMMMMMMMMMMMMMMMMMMMMMMMM",
        "cdr1": "MMMMMMIIIIIIMMMMMM",
        "fwr2": "MMMMMMMMMMMMMMMMM",
        "cdr2": "MMMMMIIIMMMMM",
        "fwr3": "MMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMM",
        "cdr3": "MMMMMMMIIIIIIMMMMMM",
        "fwr4": "MMMMMMMMMMM",
    }
    then_alignment_str = AlignmentString("".join(regions.values()))
    assert sum(map(then_alignment_str.count, ["M", "D", "N"])) == 128

    assert (
        smooth_cdr_junctions(alignment_str=given_alignment_str, chain_type=CHAIN_TYPE, scheme=SCHEME)
        == then_alignment_str
    )


def test_smooth_cdr_junctions_imgt_insertions_and_deletions():
    regions = {
        "fwr1": "MMIMDMMIMMMMMMMMMMMDMMMMMMMD",
        "cdr1": "MIMMMDMMMMMIMM",
        "fwr2": "IMIMMMMMMMMMMMMMMMM",
        "cdr2": "IMMMDMMMMMIIMI",
        "fwr3": "DMMMMMMMMDMMMMMMMMMMIMMMMMDMMMMMIMMMMMMDD",
        "cdr3": "DMIIMDMMDDMMIMDMI",
        "fwr4": "MIMMMMMIMDMMM",
    }
    given_alignment_str = AlignmentString("".join(regions.values()))
    assert sum(map(given_alignment_str.count, ["M", "D", "N"])) == 128

    regions = {
        "fwr1": "MMIMDMMIMMMMMMMMMMMDMMMMMMMD",
        "cdr1": "MMMMMMIIMMMMMM",
        "fwr2": "MIMMMMMMMMMMMMMMMM",
        "cdr2": "MMMMMIIIMMMMM",
        "fwr3": "DMMMMMMMMDMMMMMMMMMMIMMMMMDMMMMMIMMMMMMDD",
        "cdr3": "MMMMMMDMMMMMM",
        "fwr4": "MIMMMMMIMDMMM",
    }
    then_alignment_str = AlignmentString("".join(regions.values()))
    assert sum(map(then_alignment_str.count, ["M", "D", "N"])) == 128

    assert (
        smooth_cdr_junctions(alignment_str=given_alignment_str, chain_type=CHAIN_TYPE, scheme=SCHEME)
        == then_alignment_str
    )


def test_smooth_cdr_junctions_imgt_start_on_last_fwr1():
    regions = {
        "fwr1": "NNNNNNNNNNNNNNNNNNNNNNNNNM",
        "cdr1": "MMMMMMMMMMMM",
        "fwr2": "MMMMMMMMMMMMMMMMM",
        "cdr2": "MMMMMMMMMM",
        "fwr3": "MMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMM",
        "cdr3": "MMMMMMMMMMMMM",
        "fwr4": "MMMMMMMMMMM",
    }
    given_alignment_str = AlignmentString("".join(regions.values()))
    assert sum(map(given_alignment_str.count, ["M", "D", "N"])) == 128

    regions = {
        "fwr1": "NNNNNNNNNNNNNNNNNNNNNNNNNM",
        "cdr1": "MMMMMMMMMMMM",
        "fwr2": "MMMMMMMMMMMMMMMMM",
        "cdr2": "MMMMMMMMMM",
        "fwr3": "MMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMM",
        "cdr3": "MMMMMMMMMMMMM",
        "fwr4": "MMMMMMMMMMM",
    }
    then_alignment_str = AlignmentString("".join(regions.values()))
    assert sum(map(then_alignment_str.count, ["M", "D", "N"])) == 128

    assert (
        smooth_cdr_junctions(alignment_str=given_alignment_str, chain_type=CHAIN_TYPE, scheme=SCHEME)
        == then_alignment_str
    )


def test_smooth_cdr_junctions_imgt_start_on_cdr1_start():
    regions = {
        "fwr1": "NNNNNNNNNNNNNNNNNNNNNNNNNN",
        "cdr1": "MMMMMMMMMMMM",
        "fwr2": "MMMMMMMMMMMMMMMMM",
        "cdr2": "MMMMMMMMMM",
        "fwr3": "MMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMM",
        "cdr3": "MMMMMMMMMMMMM",
        "fwr4": "MMMMMMMMMMM",
    }
    given_alignment_str = AlignmentString("".join(regions.values()))
    assert sum(map(given_alignment_str.count, ["M", "D", "N"])) == 128

    regions = {
        "fwr1": "NNNNNNNNNNNNNNNNNNNNNNNNNN",
        "cdr1": "MMMMMMMMMMMM",
        "fwr2": "MMMMMMMMMMMMMMMMM",
        "cdr2": "MMMMMMMMMM",
        "fwr3": "MMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMM",
        "cdr3": "MMMMMMMMMMMMM",
        "fwr4": "MMMMMMMMMMM",
    }
    then_alignment_str = AlignmentString("".join(regions.values()))
    assert sum(map(then_alignment_str.count, ["M", "D", "N"])) == 128

    assert (
        smooth_cdr_junctions(alignment_str=given_alignment_str, chain_type=CHAIN_TYPE, scheme=SCHEME)
        == then_alignment_str
    )


def test_smooth_cdr_junctions_imgt_start_on_cdr1():
    regions = {
        "fwr1": "NNNNNNNNNNNNNNNNNNNNNNNNNN",
        "cdr1": "NNNNNMMMMMMM",
        "fwr2": "MMMMMMMMMMMMMMMMM",
        "cdr2": "MMMMMMMMMM",
        "fwr3": "MMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMM",
        "cdr3": "MMMMMMMMMMMMM",
        "fwr4": "MMMMMMMMMMM",
    }
    given_alignment_str = AlignmentString("".join(regions.values()))
    assert sum(map(given_alignment_str.count, ["M", "D", "N"])) == 128

    regions = {
        "fwr1": "NNNNNNNNNNNNNNNNNNNNNNNNNN",
        "cdr1": "MMMMDDDDDMMM",
        "fwr2": "MMMMMMMMMMMMMMMMM",
        "cdr2": "MMMMMMMMMM",
        "fwr3": "MMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMM",
        "cdr3": "MMMMMMMMMMMMM",
        "fwr4": "MMMMMMMMMMM",
    }
    then_alignment_str = AlignmentString("".join(regions.values()))
    assert sum(map(then_alignment_str.count, ["M", "D", "N"])) == 128

    assert (
        smooth_cdr_junctions(alignment_str=given_alignment_str, chain_type=CHAIN_TYPE, scheme=SCHEME)
        == then_alignment_str
    )


def test_smooth_cdr_junctions_imgt_start_on_cdr1_end():
    regions = {
        "fwr1": "NNNNNNNNNNNNNNNNNNNNNNNNNN",
        "cdr1": "NNNNNNNNNNNM",
        "fwr2": "MMMMMMMMMMMMMMMMM",
        "cdr2": "MMMMMMMMMM",
        "fwr3": "MMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMM",
        "cdr3": "MMMMMMMMMMMMM",
        "fwr4": "MMMMMMMMMMM",
    }
    given_alignment_str = AlignmentString("".join(regions.values()))
    assert sum(map(given_alignment_str.count, ["M", "D", "N"])) == 128

    regions = {
        "fwr1": "NNNNNNNNNNNNNNNNNNNNNNNNNN",
        "cdr1": "MDDDDDDDDDDD",
        "fwr2": "MMMMMMMMMMMMMMMMM",
        "cdr2": "MMMMMMMMMM",
        "fwr3": "MMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMM",
        "cdr3": "MMMMMMMMMMMMM",
        "fwr4": "MMMMMMMMMMM",
    }
    then_alignment_str = AlignmentString("".join(regions.values()))
    assert sum(map(then_alignment_str.count, ["M", "D", "N"])) == 128

    assert (
        smooth_cdr_junctions(alignment_str=given_alignment_str, chain_type=CHAIN_TYPE, scheme=SCHEME)
        == then_alignment_str
    )


def test_smooth_cdr_junctions_imgt_start_on_fwr2():
    regions = {
        "fwr1": "NNNNNNNNNNNNNNNNNNNNNNNNNN",
        "cdr1": "NNNNNNNNNNNN",
        "fwr2": "MDMMMMMMMMMMMMMMM",
        "cdr2": "MMMMMMMMMM",
        "fwr3": "MMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMM",
        "cdr3": "MMMMMMMMMMMMM",
        "fwr4": "MMMMMMMMMMM",
    }
    given_alignment_str = AlignmentString("".join(regions.values()))
    assert sum(map(given_alignment_str.count, ["M", "D", "N"])) == 128

    regions = {
        "fwr1": "NNNNNNNNNNNNNNNNNNNNNNNNNN",
        "cdr1": "NNNNNNNNNNNN",
        "fwr2": "MDMMMMMMMMMMMMMMM",
        "cdr2": "MMMMMMMMMM",
        "fwr3": "MMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMM",
        "cdr3": "MMMMMMMMMMMMM",
        "fwr4": "MMMMMMMMMMM",
    }
    then_alignment_str = AlignmentString("".join(regions.values()))
    assert sum(map(then_alignment_str.count, ["M", "D", "N"])) == 128

    assert (
        smooth_cdr_junctions(alignment_str=given_alignment_str, chain_type=CHAIN_TYPE, scheme=SCHEME)
        == then_alignment_str
    )


def test_smooth_cdr_junctions_imgt_short_fwr4():
    regions = {
        "fwr1": "MMMMMMMMMMMMMMMMMMMMMMMMMM",
        "cdr1": "MMMMMMMMMMMM",
        "fwr2": "MMMMMMMMMMMMMMMMM",
        "cdr2": "MMMMMMMMMM",
        "fwr3": "MMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMM",
        "cdr3": "MMMMMMMMMMMMM",
        "fwr4": "MMMM",
    }
    given_alignment_str = AlignmentString("".join(regions.values()))
    assert sum(map(given_alignment_str.count, ["M", "D", "N"])) == 121

    regions = {
        "fwr1": "MMMMMMMMMMMMMMMMMMMMMMMMMM",
        "cdr1": "MMMMMMMMMMMM",
        "fwr2": "MMMMMMMMMMMMMMMMM",
        "cdr2": "MMMMMMMMMM",
        "fwr3": "MMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMM",
        "cdr3": "MMMMMMMMMMMMM",
        "fwr4": "MMMM",
    }
    then_alignment_str = AlignmentString("".join(regions.values()))
    assert sum(map(then_alignment_str.count, ["M", "D", "N"])) == 121

    assert (
        smooth_cdr_junctions(alignment_str=given_alignment_str, chain_type=CHAIN_TYPE, scheme=SCHEME)
        == then_alignment_str
    )


def test_smooth_cdr_junctions_imgt_end_on_fwr4_junction():
    regions = {
        "fwr1": "MMMMMMMMMMMMMMMMMMMMMMMMMM",
        "cdr1": "MMMMMMMMMMMM",
        "fwr2": "MMMMMMMMMMMMMMMMM",
        "cdr2": "MMMMMMMMMM",
        "fwr3": "MMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMM",
        "cdr3": "MMDMMMMMMMMMM",
        "fwr4": "DM",
    }
    given_alignment_str = AlignmentString("".join(regions.values()))
    assert sum(map(given_alignment_str.count, ["M", "D", "N"])) == 119

    regions = {
        "fwr1": "MMMMMMMMMMMMMMMMMMMMMMMMMM",
        "cdr1": "MMMMMMMMMMMM",
        "fwr2": "MMMMMMMMMMMMMMMMM",
        "cdr2": "MMMMMMMMMM",
        "fwr3": "MMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMM",
        "cdr3": "MMMMMMDMMMMMM",
        "fwr4": "DM",
    }
    then_alignment_str = AlignmentString("".join(regions.values()))
    assert sum(map(then_alignment_str.count, ["M", "D", "N"])) == 119

    assert (
        smooth_cdr_junctions(alignment_str=given_alignment_str, chain_type=CHAIN_TYPE, scheme=SCHEME)
        == then_alignment_str
    )


def test_smooth_cdr_junctions_imgt_end_in_fwr4_junction():
    regions = {
        "fwr1": "MMMMMMMMMMMMMMMMMMMMMMMMMM",
        "cdr1": "MMMMMMMMMMMM",
        "fwr2": "MMMMMMMMMMMMMMMMM",
        "cdr2": "MMMMMMMMMM",
        "fwr3": "MMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMM",
        "cdr3": "MMDMMMMMMMMMM",
        "fwr4": "M",
    }
    given_alignment_str = AlignmentString("".join(regions.values()))
    assert sum(map(given_alignment_str.count, ["M", "D", "N"])) == 118

    regions = {
        "fwr1": "MMMMMMMMMMMMMMMMMMMMMMMMMM",
        "cdr1": "MMMMMMMMMMMM",
        "fwr2": "MMMMMMMMMMMMMMMMM",
        "cdr2": "MMMMMMMMMM",
        "fwr3": "MMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMM",
        "cdr3": "MMMMMMDMMMMMM",
        "fwr4": "M",
    }
    then_alignment_str = AlignmentString("".join(regions.values()))
    assert sum(map(then_alignment_str.count, ["M", "D", "N"])) == 118

    assert (
        smooth_cdr_junctions(alignment_str=given_alignment_str, chain_type=CHAIN_TYPE, scheme=SCHEME)
        == then_alignment_str
    )


def test_smooth_cdr_junctions_imgt_no_fwr4():
    regions = {
        "fwr1": "MMMMMMMMMMMMMMMMMMMMMMMMMM",
        "cdr1": "MMMMMMMMMMMM",
        "fwr2": "MMMMMMMMMMMMMMMMM",
        "cdr2": "MMMMMMMMMM",
        "fwr3": "MMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMM",
        "cdr3": "MMDMMMMMMMMMM",
        "fwr4": "",
    }
    given_alignment_str = AlignmentString("".join(regions.values()))
    assert sum(map(given_alignment_str.count, ["M", "D", "N"])) == 117

    regions = {
        "fwr1": "MMMMMMMMMMMMMMMMMMMMMMMMMM",
        "cdr1": "MMMMMMMMMMMM",
        "fwr2": "MMMMMMMMMMMMMMMMM",
        "cdr2": "MMMMMMMMMM",
        "fwr3": "MMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMM",
        "cdr3": "MMMMMMDMMMMMM",
        "fwr4": "",
    }
    then_alignment_str = AlignmentString("".join(regions.values()))
    assert sum(map(then_alignment_str.count, ["M", "D", "N"])) == 117

    assert (
        smooth_cdr_junctions(alignment_str=given_alignment_str, chain_type=CHAIN_TYPE, scheme=SCHEME)
        == then_alignment_str
    )


if __name__ == "__main__":
    test_smooth_cdr_junctions_imgt_no_fwr4()


def test_smooth_cdr_junctions_imgt_end_on_cdr3():
    regions = {
        "fwr1": "MMMMMMMMMMMMMMMMMMMMMMMMMM",
        "cdr1": "MMMMMMMMMMMM",
        "fwr2": "MMMMMMMMMMMMMMMMM",
        "cdr2": "MMMMMMMMMM",
        "fwr3": "MMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMM",
        "cdr3": "MMDMMMMMMM",
        "fwr4": "",
    }
    given_alignment_str = AlignmentString("".join(regions.values()))
    assert sum(map(given_alignment_str.count, ["M", "D", "N"])) == 114

    regions = {
        "fwr1": "MMMMMMMMMMMMMMMMMMMMMMMMMM",
        "cdr1": "MMMMMMMMMMMM",
        "fwr2": "MMMMMMMMMMMMMMMMM",
        "cdr2": "MMMMMMMMMM",
        "fwr3": "MMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMM",
        "cdr3": "MMMMMDDDDMMMM",
        "fwr4": "",
    }
    then_alignment_str = AlignmentString("".join(regions.values()))
    assert sum(map(then_alignment_str.count, ["M", "D", "N"])) == 117

    assert (
        smooth_cdr_junctions(alignment_str=given_alignment_str, chain_type=CHAIN_TYPE, scheme=SCHEME)
        == then_alignment_str
    )


def test_smooth_cdr_junctions_imgt_end_on_cdr3_before_indel_pos():
    regions = {
        "fwr1": "MMMMMMMMMMMMMMMMMMMMMMMMMM",
        "cdr1": "MMMMMMMMMMMM",
        "fwr2": "MMMMMMMMMMMMMMMMM",
        "cdr2": "MMMMMMMMMM",
        "fwr3": "MMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMM",
        "cdr3": "MMDM",
        "fwr4": "",
    }
    given_alignment_str = AlignmentString("".join(regions.values()))
    assert sum(map(given_alignment_str.count, ["M", "D", "N"])) == 108

    regions = {
        "fwr1": "MMMMMMMMMMMMMMMMMMMMMMMMMM",
        "cdr1": "MMMMMMMMMMMM",
        "fwr2": "MMMMMMMMMMMMMMMMM",
        "cdr2": "MMMMMMMMMM",
        "fwr3": "MMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMM",
        "cdr3": "MMDDDDDDDDDDM",
        "fwr4": "",
    }
    then_alignment_str = AlignmentString("".join(regions.values()))
    assert sum(map(then_alignment_str.count, ["M", "D", "N"])) == 117

    assert (
        smooth_cdr_junctions(alignment_str=given_alignment_str, chain_type=CHAIN_TYPE, scheme=SCHEME)
        == then_alignment_str
    )


def test_smooth_cdr_junctions_imgt_single_match_on_cdr3():
    regions = {
        "fwr1": "MMMMMMMMMMMMMMMMMMMMMMMMMM",
        "cdr1": "MMMMMMMMMMMM",
        "fwr2": "MMMMMMMMMMMMMMMMM",
        "cdr2": "MMMMMMMMMM",
        "fwr3": "MMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMM",
        "cdr3": "M",
        "fwr4": "",
    }
    given_alignment_str = AlignmentString("".join(regions.values()))
    assert sum(map(given_alignment_str.count, ["M", "D", "N"])) == 105

    regions = {
        "fwr1": "MMMMMMMMMMMMMMMMMMMMMMMMMM",
        "cdr1": "MMMMMMMMMMMM",
        "fwr2": "MMMMMMMMMMMMMMMMM",
        "cdr2": "MMMMMMMMMM",
        "fwr3": "MMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMM",
        "cdr3": "MDDDDDDDDDDDD",
        "fwr4": "",
    }
    then_alignment_str = AlignmentString("".join(regions.values()))
    assert sum(map(then_alignment_str.count, ["M", "D", "N"])) == 117

    assert (
        smooth_cdr_junctions(alignment_str=given_alignment_str, chain_type=CHAIN_TYPE, scheme=SCHEME)
        == then_alignment_str
    )


def test_smooth_cdr_junctions_imgt_end_on_fwr3_junction():
    regions = {
        "fwr1": "MMMMMMMMMMMMMMMMMMMMMMMMMM",
        "cdr1": "MMMMMMMMMMMM",
        "fwr2": "MMMMMMMMMMMMMMMMM",
        "cdr2": "MMMMMMMMMM",
        "fwr3": "MMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMM",
        "cdr3": "",
        "fwr4": "",
    }
    given_alignment_str = AlignmentString("".join(regions.values()))
    assert sum(map(given_alignment_str.count, ["M", "D", "N"])) == 104

    regions = {
        "fwr1": "MMMMMMMMMMMMMMMMMMMMMMMMMM",
        "cdr1": "MMMMMMMMMMMM",
        "fwr2": "MMMMMMMMMMMMMMMMM",
        "cdr2": "MMMMMMMMMM",
        "fwr3": "MMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMM",
        "cdr3": "",
        "fwr4": "",
    }
    then_alignment_str = AlignmentString("".join(regions.values()))
    assert sum(map(then_alignment_str.count, ["M", "D", "N"])) == 104

    assert (
        smooth_cdr_junctions(alignment_str=given_alignment_str, chain_type=CHAIN_TYPE, scheme=SCHEME)
        == then_alignment_str
    )


def test_smooth_cdr_junctions_imgt_end_on_fwr3_junction_last_missing():
    regions = {
        "fwr1": "MMMMMMMMMMMMMMMMMMMMMMMMMM",
        "cdr1": "MMMMMMMMMMMM",
        "fwr2": "MMMMMMMMMMMMMMMMM",
        "cdr2": "MMMMMMMMMM",
        "fwr3": "MMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMM",
        "cdr3": "",
        "fwr4": "",
    }
    given_alignment_str = AlignmentString("".join(regions.values()))
    assert sum(map(given_alignment_str.count, ["M", "D", "N"])) == 103

    regions = {
        "fwr1": "MMMMMMMMMMMMMMMMMMMMMMMMMM",
        "cdr1": "MMMMMMMMMMMM",
        "fwr2": "MMMMMMMMMMMMMMMMM",
        "cdr2": "MMMMMMMMMM",
        "fwr3": "MMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMM",
        "cdr3": "",
        "fwr4": "",
    }
    then_alignment_str = AlignmentString("".join(regions.values()))
    assert sum(map(then_alignment_str.count, ["M", "D", "N"])) == 103

    assert (
        smooth_cdr_junctions(alignment_str=given_alignment_str, chain_type=CHAIN_TYPE, scheme=SCHEME)
        == then_alignment_str
    )


def test_smooth_cdr_junctions_imgt_end_on_fwr3():
    regions = {
        "fwr1": "MMMMMMMMMMMMMMMMMMMMMMMMMM",
        "cdr1": "MMMMMMMMMMMM",
        "fwr2": "MMMMMMMMMMMMMMMMM",
        "cdr2": "MMMMMMMMMM",
        "fwr3": "MMMMMMMMMMMMMMMMMMMMMMMMMMMMMDMMM",
        "cdr3": "",
        "fwr4": "",
    }
    given_alignment_str = AlignmentString("".join(regions.values()))
    assert sum(map(given_alignment_str.count, ["M", "D", "N"])) == 98

    regions = {
        "fwr1": "MMMMMMMMMMMMMMMMMMMMMMMMMM",
        "cdr1": "MMMMMMMMMMMM",
        "fwr2": "MMMMMMMMMMMMMMMMM",
        "cdr2": "MMMMMMMMMM",
        "fwr3": "MMMMMMMMMMMMMMMMMMMMMMMMMMMMMDMMM",
        "cdr3": "",
        "fwr4": "",
    }
    then_alignment_str = AlignmentString("".join(regions.values()))
    assert sum(map(then_alignment_str.count, ["M", "D", "N"])) == 98

    assert (
        smooth_cdr_junctions(alignment_str=given_alignment_str, chain_type=CHAIN_TYPE, scheme=SCHEME)
        == then_alignment_str
    )


def test_smooth_cdr_junctions_imgt_end_on_cdr2():
    regions = {
        "fwr1": "MMMMMMMMMMMMMMMMMMMMMMMMMM",
        "cdr1": "MMMMMMMMMMMM",
        "fwr2": "MMMMMMMMMMMMMMMMM",
        "cdr2": "MMMMMMMDM",
        "fwr3": "",
        "cdr3": "",
        "fwr4": "",
    }
    given_alignment_str = AlignmentString("".join(regions.values()))
    assert sum(map(given_alignment_str.count, ["M", "D", "N"])) == 64

    regions = {
        "fwr1": "MMMMMMMMMMMMMMMMMMMMMMMMMM",
        "cdr1": "MMMMMMMMMMMM",
        "fwr2": "MMMMMMMMMMMMMMMMM",
        "cdr2": "MMMMDDMMMM",
        "fwr3": "",
        "cdr3": "",
        "fwr4": "",
    }
    then_alignment_str = AlignmentString("".join(regions.values()))
    assert sum(map(then_alignment_str.count, ["M", "D", "N"])) == 65

    assert (
        smooth_cdr_junctions(alignment_str=given_alignment_str, chain_type=CHAIN_TYPE, scheme=SCHEME)
        == then_alignment_str
    )


def test_reorder_buffer_deletions_imgt_no_matches():
    # even buffer len
    relative_indel_position = 6
    buffer = 14 * "D"

    matches = buffer.count("M")
    deletions = buffer.count("D")

    assert _reorder_cdr_deletions_imgt(relative_indel_position, matches, deletions) == buffer

    # odd buffer len
    relative_indel_position = 7
    buffer = 15 * "D"

    matches = buffer.count("M")
    deletions = buffer.count("D")

    assert _reorder_cdr_deletions_imgt(relative_indel_position, matches, deletions) == buffer


def test_reorder_buffer_deletions_imgt_single_match_on_right_junction():
    # even buffer len
    relative_indel_position = 6
    buffer = 13 * "D" + "M"

    matches = buffer.count("M")
    deletions = buffer.count("D")

    assert _reorder_cdr_deletions_imgt(relative_indel_position, matches, deletions) == buffer

    # odd buffer len
    relative_indel_position = 7
    buffer = 14 * "D" + "M"

    matches = buffer.count("M")
    deletions = buffer.count("D")

    assert _reorder_cdr_deletions_imgt(relative_indel_position, matches, deletions) == buffer


if __name__ == "__main__":
    test_reorder_buffer_deletions_imgt_single_match_on_right_junction()


def test_reorder_buffer_deletions_imgt_two_matches_on_right_junction():
    # even buffer len
    relative_indel_position = 6
    buffer = 12 * "D" + "MM"

    matches = buffer.count("M")
    deletions = buffer.count("D")

    assert _reorder_cdr_deletions_imgt(relative_indel_position, matches, deletions) == buffer

    # odd buffer len
    relative_indel_position = 7
    buffer = 13 * "D" + "MM"

    matches = buffer.count("M")
    deletions = buffer.count("D")

    assert _reorder_cdr_deletions_imgt(relative_indel_position, matches, deletions) == buffer


def test_reorder_buffer_deletions_imgt_three_matches_on_right_junction():
    # even buffer len
    relative_indel_position = 6
    buffer = 11 * "D" + "MMM"

    matches = buffer.count("M")
    deletions = buffer.count("D")

    assert _reorder_cdr_deletions_imgt(relative_indel_position, matches, deletions) == "M" + 11 * "D" + "MM"

    # odd buffer len
    relative_indel_position = 7
    buffer = 12 * "D" + "MMM"

    matches = buffer.count("M")
    deletions = buffer.count("D")

    assert _reorder_cdr_deletions_imgt(relative_indel_position, matches, deletions) == "M" + 12 * "D" + "MM"


def test_smooth_cdr_junctions_imgt_insertions_and_deletions_light_chain():
    regions = {
        "fwr1": "MMIMDMMIMMMMMMMMMMMDMMMMMMMD",
        "cdr1": "MIMMMDMMMMMIMM",
        "fwr2": "IMIMMMMMMMMMMMMMMMM",
        "cdr2": "IMMMDMMMMMIIMI",
        "fwr3": "DMMMMMMMMDMMMMMMMMMMIMMMMMDMMMMMIMMMMMMDD",
        "cdr3": "DMIIMDMMDDMMIMDMI",
        "fwr4": "MIMMMMMIMDMMM",
    }
    given_alignment_str = AlignmentString("".join(regions.values()))
    assert sum(map(given_alignment_str.count, ["M", "D", "N"])) == 128

    regions = {
        "fwr1": "MMIMDMMIMMMMMMMMMMMDMMMMMMMD",
        "cdr1": "MMMMMMIIMMMMMM",
        "fwr2": "MIMMMMMMMMMMMMMMMM",
        "cdr2": "MMMMMIIIMMMMM",
        "fwr3": "DMMMMMMMMDMMMMMMMMMMIMMMMMDMMMMMIMMMMMMDD",
        "cdr3": "MMMMMMDMMMMMM",
        "fwr4": "MIMMMMMIMDMMM",
    }
    then_alignment_str = AlignmentString("".join(regions.values()))
    assert sum(map(then_alignment_str.count, ["M", "D", "N"])) == 128

    assert (
        smooth_cdr_junctions(alignment_str=given_alignment_str, chain_type=ChainType.LIGHT, scheme=SCHEME)
        == then_alignment_str
    )


if __name__ == "__main__":
    test_smooth_cdr_junctions_imgt_no_indels()
    test_smooth_cdr_junctions_imgt_insertions_and_deletions()
    test_smooth_cdr_junctions_imgt_start_on_last_fwr1()
    test_smooth_cdr_junctions_imgt_start_on_cdr1_start()
    test_smooth_cdr_junctions_imgt_start_on_cdr1()
    test_smooth_cdr_junctions_imgt_start_on_cdr1_end()
    test_smooth_cdr_junctions_imgt_start_on_fwr2()
    test_smooth_cdr_junctions_imgt_short_fwr4()
    test_smooth_cdr_junctions_imgt_end_on_fwr4_junction()
    test_smooth_cdr_junctions_imgt_end_in_fwr4_junction()
    test_smooth_cdr_junctions_imgt_no_fwr4()
    test_smooth_cdr_junctions_imgt_end_on_cdr3()
    test_smooth_cdr_junctions_imgt_end_on_cdr3_before_indel_pos()
    test_smooth_cdr_junctions_imgt_single_match_on_cdr3()
    test_smooth_cdr_junctions_imgt_end_on_fwr3_junction()
    test_smooth_cdr_junctions_imgt_end_on_fwr3_junction_last_missing()
    test_smooth_cdr_junctions_imgt_end_on_fwr3()
    test_smooth_cdr_junctions_imgt_end_on_cdr2()
    test_smooth_cdr_junctions_imgt_insertions_and_deletions_light_chain()



from pathlib import Path
from typing import Literal, Optional

from riot_na.alignment.alignment_utils import unfold_cigar
from riot_na.config import GENE_DB_DIR
from riot_na.data.constants import AMINO_ACIDS
from riot_na.data.model import (
    AlignmentEntryAA,
    AlignmentsAA,
    AlignmentString,
    ChainType,
    Cigar,
    Locus,
    Organism,
    Scheme,
    SchemeAlignment,
)
from riot_na.data.scheme_definitions import get_legal_positions
from riot_na.data.scheme_mapping_facade import SchemeMappingFacade
from riot_na.schemes.collapse_alignment import collapse_alignment_str, collapse_ins_del
from riot_na.schemes.smooth_alignment import smooth_cdr_junctions

IMGT_REVERSE_ORDER_INSERTIONS = [33, 61, 112]


def validate_sequence(sequence):
    """
    Check whether a sequence is a protein sequence or if someone has submitted something nasty.
    """
    assert len(sequence) < 10000, "Sequence too long."
    unknown_aa = set(sequence.upper()) - AMINO_ACIDS
    assert not (unknown_aa), f"Unknown amino acid letter found in sequence: {list(unknown_aa)}"


def validate_alignment_len(aln: AlignmentEntryAA, gene: Literal["V", "J"], min_len: int):
    """
    Check whether alignment is not too short.
    """
    assert (
        aln.q_end - aln.q_start >= min_len
    ), f"{gene} sequence alignment AA too short, length: {aln.q_end - aln.q_start}"


def force_n_terminus_matches(alignment: AlignmentEntryAA) -> AlignmentEntryAA:
    # force matches on N terminus for V gene
    n_term_number_of_matches = min(alignment.q_start, alignment.t_start)
    fixed_cigar = Cigar(f"{n_term_number_of_matches}M" + alignment.cigar)

    q_start = alignment.q_start - n_term_number_of_matches
    t_start = alignment.t_start - n_term_number_of_matches

    return AlignmentEntryAA(
        target_id=alignment.target_id,
        alignment_score=None,
        seq_identity=None,
        e_value=None,
        q_start=q_start,
        q_end=alignment.q_end,
        t_start=t_start,
        t_end=alignment.t_end,
        cigar=fixed_cigar,
        species=alignment.species,
        locus=alignment.locus,
        q_seq=alignment.q_seq,
        t_seq=alignment.t_seq,
    )


def force_c_terminus_matches(
    query_sequence: str, target_sequence: str, alignment: AlignmentEntryAA
) -> AlignmentEntryAA:
    # force matches on C terminus for J gene
    # c_term_number_of_matches = min(len(query_sequence) - (alignment.q_end -alignment.q_start), len(target_sequence) - alignment.t_end)
    c_term_number_of_matches = min(len(query_sequence) - (alignment.q_end), len(target_sequence) - alignment.t_end)
    fixed_cigar = Cigar(alignment.cigar + f"{c_term_number_of_matches}M")

    q_end = alignment.q_end + c_term_number_of_matches
    t_end = alignment.t_end + c_term_number_of_matches

    return AlignmentEntryAA(
        target_id=alignment.target_id,
        alignment_score=None,
        seq_identity=None,
        e_value=None,
        q_start=alignment.q_start,
        q_end=q_end,
        t_start=alignment.t_start,
        t_end=t_end,
        cigar=fixed_cigar,
        species=alignment.species,
        locus=alignment.locus,
        q_seq=alignment.q_seq,
        t_seq=alignment.t_seq,
    )


def force_n_terminus_del_ins(aln: AlignmentEntryAA, query_gene_alignment_str: AlignmentString) -> AlignmentString:
    # add deletions and insertions on N terminus of alignment so the cigar covers whole gene
    # this is needed to infer the last scheme position of V gene and first position of J gene
    if aln.t_start > 0:
        query_gene_alignment_str = AlignmentString(aln.t_start * "N" + query_gene_alignment_str)

    if aln.q_start > 0:
        query_gene_alignment_str = AlignmentString(aln.q_start * "S" + query_gene_alignment_str)
    return query_gene_alignment_str


def _merge_cigars(
    query_gene_alignment_str: AlignmentString,
    gene_scheme_alignment_str: AlignmentString,
    query_start: int = 0,
    target_start: int = 0,
) -> AlignmentString:
    # I M -> I: it1++ -ok
    # I D -> M:it1++, it2++ -ok
    # I I -> I: it2++ !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    # M M ->M:it1++, it2++ -ok
    # M D -> D: it2++ -ok
    # M I -> I: it1++, it2++ !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    # D D -> D: it1++ !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    # D M -> : D : it1++, it2++
    # D I -> X: it1++, it2++

    #     'SSSSSSSSSSSSSNNNMMMMMMMM'
    # 'NNNNNNNNNNNNNNNMMMDMMMMMMMMM'

    query_gene_pos = query_start + target_start
    gene_scheme_pos = target_start
    res_cigar = ""

    while query_gene_pos < len(query_gene_alignment_str) and gene_scheme_pos < len(gene_scheme_alignment_str):
        query_gene_op = query_gene_alignment_str[query_gene_pos]
        gene_scheme_op = gene_scheme_alignment_str[gene_scheme_pos]

        match (query_gene_op, gene_scheme_op):
            case ("I", "M"):
                query_gene_pos = query_gene_pos + 1
                res_cigar = res_cigar + "I"
            case ("I", "D"):
                query_gene_pos = query_gene_pos + 1
                gene_scheme_pos = gene_scheme_pos + 1
                res_cigar = res_cigar + "M"
            case ("I", "I"):
                query_gene_pos = query_gene_pos + 1
                res_cigar = res_cigar + "I"
            case ("M", "M"):
                query_gene_pos = query_gene_pos + 1
                gene_scheme_pos = gene_scheme_pos + 1
                res_cigar = res_cigar + "M"
            case ("M", "D"):
                gene_scheme_pos = gene_scheme_pos + 1
                res_cigar = res_cigar + "D"
            case ("M", "I"):
                query_gene_pos = query_gene_pos + 1
                gene_scheme_pos = gene_scheme_pos + 1
                res_cigar = res_cigar + "I"
            case ("D", "D"):
                gene_scheme_pos = gene_scheme_pos + 1
                res_cigar = res_cigar + "D"
            case ("D", "M"):
                query_gene_pos = query_gene_pos + 1
                gene_scheme_pos = gene_scheme_pos + 1
                res_cigar = res_cigar + "D"
            case ("D", "I"):
                query_gene_pos = query_gene_pos + 1
                gene_scheme_pos = gene_scheme_pos + 1
            case ("N", "M"):
                query_gene_pos = query_gene_pos + 1
                gene_scheme_pos = gene_scheme_pos + 1
                res_cigar = res_cigar + "N"
            case ("N", "D"):
                gene_scheme_pos = gene_scheme_pos + 1
                res_cigar = res_cigar + "N"
            case ("N", "I"):
                query_gene_pos = query_gene_pos + 1
                gene_scheme_pos = gene_scheme_pos + 1
            case ("S", "M"):
                query_gene_pos = query_gene_pos + 1
                res_cigar = res_cigar + "S"
            case ("S", "I"):
                query_gene_pos = query_gene_pos + 1
                res_cigar = res_cigar + "S"
            case ("S", "D"):
                query_gene_pos = query_gene_pos + 1
                gene_scheme_pos = gene_scheme_pos + 1

    while query_gene_pos < len(query_gene_alignment_str):
        query_gene_op = query_gene_alignment_str[query_gene_pos]
        assert query_gene_op not in {"M", "D"}, "Query gene alignment is longer than gene scheme alignment"
        res_cigar += query_gene_op
        query_gene_pos = query_gene_pos + 1

    return AlignmentString(res_cigar)


def infer_last_scheme_position_aligned(query_scheme_aln_str: AlignmentString) -> int:
    # query =   "CAAAAAAAAAAAA"
    # gene =    "BAAAAAAABC"
    # aln =    "SNMMMMMMM"
    # g_sch =   "MMMDMMMMMMM"
    # q_sch =  "SNMMDMMMM"

    # scheme space:
    # query =  "C-AA AAAAAAAAAA"
    # gene =   "-BAA AAAAABC"
    # aln =    "SNMM MMMMM"
    # g_sch =   "MMMDMMMMMMM"
    # q_sch =  "IDMMDMMMMM"
    # last_pos= "        |"  # 9 (1-based)

    matches = query_scheme_aln_str.count("M")
    deletions = query_scheme_aln_str.count("D") + query_scheme_aln_str.count("N")

    last_position_sch = matches + deletions  # 1-based
    return last_position_sch


def merge_v_j(
    v_query_scheme_cigar: AlignmentString,
    j_query_scheme_cigar: AlignmentString,
    j_gene_start_on_scheme: int,
) -> AlignmentString:
    last_v_match_pos = v_query_scheme_cigar.rfind("M")
    first_j_match_pos = j_query_scheme_cigar.find("M")

    last_aligned_v_position_on_scheme = infer_last_scheme_position_aligned(
        AlignmentString(v_query_scheme_cigar[: last_v_match_pos + 1])
    )

    # remaining part of alignmetns that does not match to the scheme
    middle_v_scheme_cigar = v_query_scheme_cigar[last_v_match_pos + 1 :]
    middle_j_scheme_cigar = j_query_scheme_cigar[:first_j_match_pos]

    scheme_deletions = (
        j_gene_start_on_scheme
        + middle_j_scheme_cigar.count("N")
        + middle_j_scheme_cigar.count("D")
        - last_aligned_v_position_on_scheme
        - 1
    )
    query_insertions = (
        middle_v_scheme_cigar.count("I") + middle_j_scheme_cigar.count("S") + middle_j_scheme_cigar.count("I")
    )
    middle_cigar = collapse_ins_del(scheme_deletions * ["D"] + query_insertions * ["I"])

    final_cigar = AlignmentString(
        f"{v_query_scheme_cigar[:last_v_match_pos + 1]}{middle_cigar}{j_query_scheme_cigar[first_j_match_pos:]}"
    )
    return final_cigar


IntermediateNumbering = dict[int, list[tuple[str, int]]]


def produce_numbering(
    query: str, alignment_str: AlignmentString, query_start: Optional[int] = None
) -> IntermediateNumbering:
    scheme_position = 0
    query_position = query_start or 0
    insertion_counter = 0

    numbering: IntermediateNumbering = {}
    for op in alignment_str:
        if op == "M":
            insertion_counter = 0
            query_position = query_position + 1
            scheme_position = scheme_position + 1
            position_residues = numbering.get(scheme_position, [])
            position_residues.append((query[query_position - 1], insertion_counter))
            numbering[scheme_position] = position_residues
        elif op in {"D", "N"}:
            insertion_counter = 0
            scheme_position = scheme_position + 1
        elif op == "I":
            insertion_counter = insertion_counter + 1
            query_position = query_position + 1
            position_residues = numbering.get(scheme_position, [])
            position_residues.append((query[query_position - 1], insertion_counter))
            numbering[scheme_position] = position_residues

    return numbering


def scheme_alignment(
    v_aligment: AlignmentEntryAA,
    v_gene_scheme_alignment_str: AlignmentString,
    j_aligment: Optional[AlignmentEntryAA],
    j_gene_scheme_alignment_str: Optional[AlignmentString],
    j_gene_start_on_scheme: Optional[int],
    scheme: Scheme,
    extend_alignment: bool = True,
) -> SchemeAlignment:
    fixed_v_aln = v_aligment
    if extend_alignment:
        fixed_v_aln = force_n_terminus_matches(v_aligment)

    v_query_gene_alignment_str = unfold_cigar(fixed_v_aln.cigar)
    v_query_gene_alignment_str = collapse_alignment_str(v_query_gene_alignment_str)
    v_query_gene_alignment_str = force_n_terminus_del_ins(fixed_v_aln, v_query_gene_alignment_str)

    v_query_scheme_alignment_str = _merge_cigars(v_query_gene_alignment_str, v_gene_scheme_alignment_str)
    query_scheme_alignment_str = v_query_scheme_alignment_str

    if j_aligment:
        assert j_gene_scheme_alignment_str is not None
        assert j_gene_start_on_scheme is not None
        assert j_aligment.t_seq is not None

        if j_aligment.q_end - j_aligment.q_start > 5:
            masked = v_aligment.q_seq[v_aligment.q_end :]

            # masked = v_aligment.q_seq[v_aligment.q_end - v_aligment.q_start :]
            if extend_alignment:
                j_aligment = force_c_terminus_matches(masked, j_aligment.t_seq, j_aligment)

            j_query_gene_alignment_str = unfold_cigar(j_aligment.cigar)
            j_query_gene_alignment_str = collapse_alignment_str(j_query_gene_alignment_str)
            j_query_gene_alignment_str = force_n_terminus_del_ins(j_aligment, j_query_gene_alignment_str)

            j_query_scheme_alignment_str = _merge_cigars(j_query_gene_alignment_str, j_gene_scheme_alignment_str)

            query_scheme_alignment_str = merge_v_j(
                v_query_scheme_cigar=query_scheme_alignment_str,
                j_query_scheme_cigar=j_query_scheme_alignment_str,
                j_gene_start_on_scheme=j_gene_start_on_scheme,
            )

    chain_type = ChainType.HEAVY if v_aligment.locus == Locus.IGH else ChainType.LIGHT
    collapsed_query_scheme_alignment_str = collapse_alignment_str(query_scheme_alignment_str, ordered=True)
    smoothed_collapsed_query_scheme_alignment_str = smooth_cdr_junctions(
        collapsed_query_scheme_alignment_str, chain_type, scheme
    )

    return SchemeAlignment(
        target_id=scheme.value,
        q_start=fixed_v_aln.q_start,
        q_end=j_aligment.q_end + fixed_v_aln.q_end if j_aligment else fixed_v_aln.q_end,
        alignment_str=smoothed_collapsed_query_scheme_alignment_str,
    )


def get_scheme_residue_mapping(numbering: IntermediateNumbering) -> dict[str, str]:
    result: dict[str, str] = {}

    for scheme_id, residues in numbering.items():
        for residue, insertion in residues:
            scheme_id_str = str(scheme_id) if not insertion else f"{scheme_id}.{insertion}"
            result[scheme_id_str] = residue

    return result


def sort_imgt_numbering(numbering: IntermediateNumbering) -> IntermediateNumbering:
    for pos in IMGT_REVERSE_ORDER_INSERTIONS:
        if pos in numbering:
            numbering[pos] = sorted(numbering[pos], key=lambda x: x[1], reverse=True)
    return numbering


def get_positional_scheme_mapping(scheme_residue_mapping: dict[str, str], query_start: int = 0) -> dict[int, str]:
    return dict(enumerate(scheme_residue_mapping.keys(), query_start))


def _fix_imgt_cdr_numbering(
    numbering: IntermediateNumbering,
    insertion_position: int,
    swap_priority: bool = False,
) -> IntermediateNumbering:
    insertions = numbering.get(insertion_position, [])
    if len(insertions) <= 1:
        return numbering

    new_insertions = numbering.get(insertion_position + 1, [])
    assert len(new_insertions) == 1

    new_insertion_counter = 1
    while True:
        len_diff = len(new_insertions) - len(insertions) if swap_priority else len(insertions) - len(new_insertions)
        if len_diff in [0, 1]:
            break
        residue, _ = insertions.pop()
        new_insertions.append((residue, new_insertion_counter))
        new_insertion_counter += 1

    numbering[insertion_position] = insertions
    numbering[insertion_position + 1] = new_insertions
    return numbering


def fix_imgt_cdrs_numbering(numbering: IntermediateNumbering) -> IntermediateNumbering:
    # CDR1
    numbering = _fix_imgt_cdr_numbering(numbering, insertion_position=32)
    # CDR2
    numbering = _fix_imgt_cdr_numbering(numbering, insertion_position=60)
    # CDR3
    numbering = _fix_imgt_cdr_numbering(numbering, insertion_position=111, swap_priority=True)
    return numbering


def _get_j_gene_start_on_scheme(j_gene_scheme_alignment: AlignmentString, chain_type: ChainType, scheme: Scheme) -> int:
    legal_positions = get_legal_positions(chain_type, scheme)
    j_start_on_scheme = legal_positions - j_gene_scheme_alignment.count("M") - j_gene_scheme_alignment.count("D") + 1
    return j_start_on_scheme


class SchemeAligner:
    def __init__(
        self,
        allowed_species: Optional[list[Organism]] = None,
        db_dir: Path = GENE_DB_DIR,
    ):
        self.scheme_mapping_facades: dict[Scheme, SchemeMappingFacade] = {}

        if not allowed_species:
            allowed_species = [Organism.HOMO_SAPIENS, Organism.MUS_MUSCULUS]

        for scheme in Scheme:
            self.scheme_mapping_facades[scheme] = SchemeMappingFacade(scheme, allowed_species, db_dir)

    def align_to_scheme(
        self, aa_alignments: AlignmentsAA, scheme: Scheme, extend_alignment: bool = True
    ) -> Optional[SchemeAlignment]:
        if not aa_alignments.v:
            return None

        organism = aa_alignments.v.species
        v_gene_scheme_alignment_str = self.scheme_mapping_facades[scheme].get_mapping(
            organism, aa_alignments.v.target_id
        )
        j_gene_scheme_alignment_str = (
            self.scheme_mapping_facades[scheme].get_mapping(organism, aa_alignments.j.target_id)
            if aa_alignments.j
            else None
        )

        locus = aa_alignments.v.locus
        chain_type = ChainType.from_locus(locus)

        j_start_on_scheme = (
            _get_j_gene_start_on_scheme(j_gene_scheme_alignment_str, chain_type, scheme)
            if j_gene_scheme_alignment_str
            else None
        )

        sch_alignment = scheme_alignment(
            aa_alignments.v,
            v_gene_scheme_alignment_str,
            aa_alignments.j,
            j_gene_scheme_alignment_str,
            j_start_on_scheme,
            scheme=scheme,
            extend_alignment=extend_alignment,
        )

        return sch_alignment


if __name__ == "__main__":
    import json
    from dataclasses import asdict

    sch_aligner = SchemeAligner(allowed_species=None)

    v_almt = AlignmentEntryAA(
        target_id="IGHV4-59*01",
        alignment_score=None,
        seq_identity=None,
        e_value=None,
        q_start=6,
        q_end=82,
        q_seq="LRRPCPTCSISDGSISSYYWNWIRQSPGKGLEWIGHIHYSGSTHYNPSLQSRVSISIDTSKNHFSLKLRSVTAVDTAVYYCARWGHFDTSGYFVVDYWGQGTLVTVSS",
        t_start=20,
        t_end=96,
        t_seq="QVQLQESGPGLVKPSETLSLTCTVSGGSISSYYWSWIRQPPGKGLEWIGYIYYSGSTNYNPSLKSRVTISVDTSKNQFSLKLSSVTAADTAVYYCAR",
        cigar=Cigar("76M"),
        locus=Locus.IGH,
        species=Organism.HOMO_SAPIENS,
    )

    j_almt = AlignmentEntryAA(
        target_id="IGHJ4*02",
        alignment_score=None,
        seq_identity=None,
        e_value=None,
        q_start=16,
        q_end=23,
        q_seq="RWGHFDTSGYFVVDYWGQGTLVTVSS",
        t_start=5,
        t_end=12,
        t_seq="YFDYWGQGTLVTVSS",
        cigar=Cigar("7M"),
        species=Organism.HOMO_SAPIENS,
        locus=Locus.IGH,
    )

    aa_almts = AlignmentsAA(v=v_almt, j=j_almt)
    scheme_aln = sch_aligner.align_to_scheme(aa_almts, Scheme.MARTIN)

    if scheme_aln:
        print(json.dumps(asdict(scheme_aln)))



import base64


def base64_encode(val: str) -> str:
    return base64.b64encode(val.encode("utf-8")).decode("utf-8")


def base64_decode(val: str) -> str:
    return base64.b64decode(val.encode("utf-8")).decode("utf-8")



from pathlib import Path
from typing import Iterable, Type

from Bio import SeqIO

from riot_na.common.airr_csv_writer import AirrRearrangementEntryWriter
from riot_na.data.model import AirrRearrangementEntry_co  # type: ignore


def read_fasta(path: Path) -> tuple[list[str], list[str]]:
    headers = []
    sequences = []
    for record in SeqIO.parse(path, "fasta"):
        headers.append(record.description)
        sequences.append(str(record.seq))
    return headers, sequences


def count_fasta_records(input_fasta_path: Path, input_format: str = "fasta") -> int:
    total = 0
    start_token = ">" if input_format == "fasta" else "@"
    with open(input_fasta_path) as input_file:
        for line in input_file:
            if line.startswith(start_token):
                total += 1
    return total


def write_airr_iter_to_csv(
    output_file_path: Path, cls: Type[AirrRearrangementEntry_co], airr_iter: Iterable[AirrRearrangementEntry_co]
):
    output_file_path.parent.mkdir(exist_ok=True, parents=True)
    with open(output_file_path, "w") as output:
        AirrRearrangementEntryWriter(output, cls).write(airr_iter)



from skbio.alignment import StripedSmithWaterman  # type: ignore

from riot_na.alignment.skbio_alignment import align


def test_align():
    # given
    query = "CTATACTACTATGGTTCGGGGAGTTATTATAGCCTTGACTACTGGGGCCAGGGAACCCTGGTCACCGTCTCCTCAGGGAGTGCATCCGCCCCAACCTCGT"
    target = "ACTACTTTGACTACTGGGGCCAGGGAACCCTGGTCACCGTCTCCTCAG"

    # when
    aligner = StripedSmithWaterman(query)
    alignment = align(aligner, "IGHJ4*02", target, 100, query)

    # then

    assert query[alignment.q_start : alignment.q_end] == "TTGACTACTGGGGCCAGGGAACCCTGGTCACCGTCTCCTCAG"


if __name__ == "__main__":
    test_align()



from functools import cache
from typing import Optional

from riot_na.common.assert_never import assert_never
from riot_na.data.model import ChainType, Locus, Scheme, ShortRegion
from riot_na.data.scheme_definitions import (
    CHOTHIA_REGIONS,
    IMGT_REGIONS,
    KABAT_REGIONS,
    MARTIN_REGIONS,
    ChainRegions,
    Regions,
)


class UnknownPosition(Exception):
    pass


@cache
def get_regions_by_scheme_and_chain(scheme: Scheme, chain: ChainType) -> ChainRegions:
    """
    Get the region of a given position in a specified scheme entry.
    """
    match scheme:
        case Scheme.KABAT:
            match chain:
                case ChainType.HEAVY:
                    return KABAT_REGIONS[chain]
                case ChainType.LIGHT:
                    return KABAT_REGIONS[chain]
            assert_never(chain)
        case Scheme.IMGT:
            match chain:
                case ChainType.HEAVY:
                    return IMGT_REGIONS[chain]
                case ChainType.LIGHT:
                    return IMGT_REGIONS[chain]
            assert_never(chain)
        case Scheme.CHOTHIA:
            match chain:
                case ChainType.HEAVY:
                    return CHOTHIA_REGIONS[chain]
                case ChainType.LIGHT:
                    return CHOTHIA_REGIONS[chain]
            assert_never(chain)
        case Scheme.MARTIN:
            match chain:
                case ChainType.HEAVY:
                    return MARTIN_REGIONS[chain]
                case ChainType.LIGHT:
                    return MARTIN_REGIONS[chain]
            assert_never(chain)
    assert_never(scheme)


def position_to_region(regions: ChainRegions, position: int) -> Optional[ShortRegion]:
    if regions[ShortRegion.FW1]["min"] <= position <= regions[ShortRegion.FW1]["max"]:
        return ShortRegion.FW1
    if regions[ShortRegion.CDR1]["min"] <= position <= regions[ShortRegion.CDR1]["max"]:
        return ShortRegion.CDR1
    if regions[ShortRegion.FW2]["min"] <= position <= regions[ShortRegion.FW2]["max"]:
        return ShortRegion.FW2
    if regions[ShortRegion.CDR2]["min"] <= position <= regions[ShortRegion.CDR2]["max"]:
        return ShortRegion.CDR2
    if regions[ShortRegion.FW3]["min"] <= position <= regions[ShortRegion.FW3]["max"]:
        return ShortRegion.FW3
    if regions[ShortRegion.CDR3]["min"] <= position <= regions[ShortRegion.CDR3]["max"]:
        return ShortRegion.CDR3
    if regions[ShortRegion.FW4]["min"] <= position <= regions[ShortRegion.FW4]["max"]:
        return ShortRegion.FW4

    return None


def get_region(scheme_position: int, scheme: Scheme, locus: Locus) -> ShortRegion:
    """
    Based on scheme position, get the region
    """
    chain_type = ChainType.from_locus(locus)

    region_mapping = get_regions_by_scheme_and_chain(scheme, chain_type)

    region = position_to_region(region_mapping, scheme_position)

    if region:
        return region

    raise UnknownPosition(f"Unknown position {scheme_position} for scheme {scheme} chain {chain_type}")


def get_regions_definitions() -> dict[Scheme, Regions]:
    return {
        Scheme.IMGT: IMGT_REGIONS,
        Scheme.KABAT: KABAT_REGIONS,
        Scheme.CHOTHIA: CHOTHIA_REGIONS,
        Scheme.MARTIN: CHOTHIA_REGIONS,
    }


@cache
def get_cdr_ranges(scheme: Scheme, chain_type: ChainType) -> list[tuple[int, int]]:
    def _get_cdrs(regions: ChainRegions):
        return [
            (regions[region]["min"], regions[region]["max"])
            for region in [ShortRegion.CDR1, ShortRegion.CDR2, ShortRegion.CDR3]
        ]

    match (scheme, chain_type):
        case Scheme.IMGT, ChainType.HEAVY:
            return _get_cdrs(IMGT_REGIONS[chain_type])
        case Scheme.IMGT, ChainType.LIGHT:
            return _get_cdrs(IMGT_REGIONS[chain_type])
        case Scheme.KABAT, ChainType.HEAVY:
            return _get_cdrs(KABAT_REGIONS[chain_type])
        case Scheme.KABAT, ChainType.LIGHT:
            return _get_cdrs(KABAT_REGIONS[chain_type])
        case Scheme.CHOTHIA, ChainType.HEAVY:
            return _get_cdrs(CHOTHIA_REGIONS[chain_type])
        case Scheme.CHOTHIA, ChainType.LIGHT:
            return _get_cdrs(CHOTHIA_REGIONS[chain_type])
        case Scheme.MARTIN, ChainType.HEAVY:
            return _get_cdrs(MARTIN_REGIONS[chain_type])
        case Scheme.MARTIN, ChainType.LIGHT:
            return _get_cdrs(MARTIN_REGIONS[chain_type])

    raise ValueError(f"Unknown scheme {scheme} or chain {chain_type}")


if __name__ == "__main__":
    print(list(get_cdr_ranges(Scheme.IMGT, ChainType.HEAVY)))



import itertools
import multiprocessing as mp
from pathlib import Path
from typing import Optional

from Bio import SeqIO
from psutil import cpu_count
from tqdm import tqdm

from riot_na.api.riot_numbering import create_riot_aa, create_riot_nt
from riot_na.common.io import count_fasta_records, write_airr_iter_to_csv
from riot_na.config import GENE_DB_DIR
from riot_na.data.model import (
    AirrRearrangementEntryAA,
    AirrRearrangementEntryNT,
    InputType,
    Organism,
    Scheme,
)


class _WorkerNT:
    def __init__(
        self,
        allowed_species: Optional[list[Organism]] = None,
        scheme: Scheme = Scheme.IMGT,
        db_dir: Path = GENE_DB_DIR,
    ) -> None:
        self.numbering = create_riot_nt(allowed_species=allowed_species, db_dir=db_dir)
        self.scheme = scheme

    def __call__(self, fasta_record: SeqIO.SeqRecord) -> AirrRearrangementEntryNT:
        try:
            res = self.numbering.run_on_sequence(
                fasta_record.description,
                str(fasta_record.seq),
                self.scheme,
            )
        except Exception as exc:
            print(fasta_record.description)
            print(self.scheme)
            print(f"seq: {fasta_record.seq}")
            raise exc

        return res


class _WorkerAA:
    def __init__(
        self,
        allowed_species: Optional[list[Organism]] = None,
        scheme: Scheme = Scheme.IMGT,
        db_dir: Path = GENE_DB_DIR,
    ) -> None:
        self.numbering = create_riot_aa(allowed_species=allowed_species, db_dir=db_dir)
        self.scheme = scheme

    def __call__(self, fasta_record: SeqIO.SeqRecord) -> AirrRearrangementEntryAA:
        try:
            res = self.numbering.run_on_sequence(
                fasta_record.description,
                str(fasta_record.seq),
                self.scheme,
            )
        except Exception as exc:
            print(fasta_record.description)
            print(self.scheme)
            print(f"seq: {fasta_record.seq}")
            raise exc

        return res


def _worker_initializer(
    allowed_species: Optional[list[Organism]] = None,
    scheme: Scheme = Scheme.IMGT,
    db_dir: Path = GENE_DB_DIR,
    input_type: InputType = InputType.NT,
):
    global worker  # pylint: disable=global-statement,global-variable-undefined
    if input_type == InputType.NT:
        worker = _WorkerNT(allowed_species, scheme, db_dir)  # type: ignore
    else:
        worker = _WorkerAA(allowed_species, scheme, db_dir)  # type: ignore


def _worker_call(*args, **kwds):
    return worker(*args, **kwds)  # type: ignore


def run_on_file_mp(
    db_dir: Path,
    input_fasta_path: Path,
    result_path: Path,
    n_processes: int = cpu_count(logical=False),
    input_format: str = "fasta",
    scheme: Scheme = Scheme.IMGT,
    allowed_species: Optional[list[Organism]] = None,
    input_type: InputType = InputType.NT,
    limit: Optional[int] = None,
):
    with mp.Pool(
        processes=n_processes, initializer=_worker_initializer, initargs=(allowed_species, scheme, db_dir, input_type)
    ) as pool:
        result_iter = tqdm(
            pool.imap(_worker_call, itertools.islice(SeqIO.parse(input_fasta_path, input_format), limit)),
            total=count_fasta_records(input_fasta_path, input_format=input_format),
        )

        record_type = AirrRearrangementEntryNT if input_type == InputType.NT else AirrRearrangementEntryAA
        write_airr_iter_to_csv(result_path, record_type, result_iter)



from enum import Enum
from functools import cache, partial
from typing import Callable, Literal, Optional

from riot_na.alignment.alignment_utils import translate, unfold_cigar
from riot_na.data.model import (
    AirrRearrangementEntry_co,
    AirrRearrangementEntryAA,
    AirrRearrangementEntryNT,
    ChainType,
    EntryValidationResults,
    Locus,
    Scheme,
)
from riot_na.data.scheme_definitions import get_indels_positions

Region = Literal["fwr1", "cdr1", "fwr2", "cdr2", "fwr3", "cdr3", "fwr4"]


class Gene(Enum):
    V = "v_"
    J = "j_"
    C = "c_"
    _ = ""


REGIONS: list[Region] = ["fwr1", "cdr1", "fwr2", "cdr2", "fwr3", "cdr3", "fwr4"]
REGION_TO_GENE: dict[Region, dict[Literal["start", "end"], Gene]] = {
    "fwr1": {"start": Gene.V, "end": Gene.V},
    "cdr1": {"start": Gene.V, "end": Gene.V},
    "fwr2": {"start": Gene.V, "end": Gene.V},
    "cdr2": {"start": Gene.V, "end": Gene.V},
    "fwr3": {"start": Gene.V, "end": Gene.V},
    "cdr3": {"start": Gene.V, "end": Gene.J},
    "fwr4": {"start": Gene.J, "end": Gene.J},
}

GENE_TO_REGIONS_START_END: dict[Gene, list[str]] = {
    Gene.V: [
        "fwr1_start",
        "fwr1_end",
        "cdr1_start",
        "cdr1_end",
        "fwr2_start",
        "fwr2_end",
        "cdr2_start",
        "cdr2_end",
        "fwr3_start",
        "fwr3_end",
        "cdr3_start",
    ],
    Gene.J: ["cdr3_end", "fwr4_start", "fwr4_end"],
}

IMGT_CONSERVED_RESIDUES_HEAVY = {
    "23": "C",
    "41": "W",
    "104": "C",
    "118_heavy": "W",
}
IMGT_CONSERVED_RESIDUES_LIGHT = {
    "23": "C",
    "41": "W",
    "104": "C",
    "118_light": "F",
}


class SeqType(Enum):
    NT = ""
    AA = "_aa"


KABAT_LIGHT_FRAMEWORK_INSERTIONS = {"106.1"}
KABAT_HEAVY_FRAMEWORK_INSERTIONS = {"82.1", "82.2", "82.3"}

CHOTHIA_LIGHT_FRAMEWORK_INSERTIONS = {"106.1"}
CHOTHIA_HEAVY_FRAMEWORK_INSERTIONS = {"82.1", "82.2", "82.3"}

MARTIN_LIGHT_FRAMEWORK_INSERTIONS = {"40.1", "68.1", "68.2", "68.3", "68.4", "68.5", "68.6", "68.7", "68.8", "107.1"}
MARTIN_HEAVY_FRAMEWORK_INSERTIONS = {"8.1", "8.2", "8.3", "8.4", "72.1", "72.2", "72.3"}


@cache
def get_framework_insertion_positions(scheme: Scheme, chain_type: ChainType) -> set[str]:
    # this is scheme specific
    match (scheme, chain_type):
        case Scheme.IMGT, _:
            return set()
        case Scheme.KABAT, ChainType.HEAVY:
            return KABAT_HEAVY_FRAMEWORK_INSERTIONS
        case Scheme.KABAT, ChainType.LIGHT:
            return KABAT_LIGHT_FRAMEWORK_INSERTIONS
        case Scheme.CHOTHIA, ChainType.HEAVY:
            return CHOTHIA_HEAVY_FRAMEWORK_INSERTIONS
        case Scheme.CHOTHIA, ChainType.LIGHT:
            return CHOTHIA_LIGHT_FRAMEWORK_INSERTIONS
        case Scheme.MARTIN, ChainType.HEAVY:
            return MARTIN_HEAVY_FRAMEWORK_INSERTIONS
        case Scheme.MARTIN, ChainType.LIGHT:
            return MARTIN_LIGHT_FRAMEWORK_INSERTIONS

    raise ValueError(f"Unknown scheme {scheme}")


def _getattr(rearrangement: AirrRearrangementEntry_co, attr: str, seq_type: SeqType = SeqType.NT, gene: Gene = Gene._):
    return getattr(rearrangement, f"{gene.value}{attr}{seq_type.value}")


def validate_regions_in_aligned_sequence(
    rearrangement: AirrRearrangementEntry_co, seq_type: SeqType = SeqType.NT
) -> Optional[bool]:
    concatenated_regions = ""
    for region in REGIONS:
        if not _getattr(rearrangement, region, seq_type):
            continue
        concatenated_regions += _getattr(rearrangement, region, seq_type)

    sequence_alignment = _getattr(rearrangement, "sequence_alignment", seq_type)
    if (not sequence_alignment) or (not concatenated_regions):
        return None
    return concatenated_regions in sequence_alignment


def validate_translated_regions_in_aligned_sequence_aa(rearrangement: AirrRearrangementEntryNT) -> Optional[bool]:
    concatenated_regions = ""
    for region in REGIONS:
        if not _getattr(rearrangement, region):
            continue
        concatenated_regions += _getattr(rearrangement, region)

    sequence_alignment = rearrangement.sequence_alignment_aa
    v_frame = rearrangement.v_frame
    if (not sequence_alignment) or (not concatenated_regions) or (v_frame is None):
        return None
    translated_regions = translate(concatenated_regions, 0)
    return translated_regions in sequence_alignment


def validate_consecutive_offsets(
    rearrangement: AirrRearrangementEntry_co, seq_type: SeqType = SeqType.NT
) -> Optional[bool]:
    consecutive_offsets = True

    for i in range(len(REGIONS) - 1):
        prev_region_end = _getattr(rearrangement, f"{REGIONS[i]}_end", seq_type)
        next_region_start = _getattr(rearrangement, f"{REGIONS[i+1]}_start", seq_type)
        if (
            prev_region_end is not None
            and prev_region_end > 0
            and next_region_start is not None
            and next_region_start > 0
        ):
            consecutive_offsets = consecutive_offsets and next_region_start == prev_region_end + 1

    start = 0
    end = 0
    for region in REGIONS:
        region_start = _getattr(rearrangement, f"{region}_start", seq_type)
        positive_region_start = region_start is not None and region_start > 0
        if positive_region_start:
            consecutive_offsets = consecutive_offsets and region_start > start and region_start > end

        region_end = _getattr(rearrangement, f"{region}_end", seq_type)
        positive_region_end = region_end is not None and region_end > 0
        if positive_region_end:
            consecutive_offsets = consecutive_offsets and region_end > start and region_end > end
            if positive_region_start:
                consecutive_offsets = consecutive_offsets and region_end >= region_start

        if positive_region_start:
            start = region_start
        if positive_region_end:
            end = region_end

    return consecutive_offsets


def validate_no_negative_offsets_inside_gene_alignment(
    rearrangement: AirrRearrangementEntry_co, gene: Gene, seq_type: SeqType = SeqType.NT
) -> Optional[bool]:
    regions_with_positive_start_end = [
        region
        for region in GENE_TO_REGIONS_START_END[gene]
        if _getattr(rearrangement, region, seq_type) is not None and _getattr(rearrangement, region, seq_type) > 0
    ]
    if "_".join(regions_with_positive_start_end) in "_".join(GENE_TO_REGIONS_START_END[gene]):
        return True
    return False


def validate_correct_region_offsets(
    rearrangement: AirrRearrangementEntry_co, seq_type: SeqType = SeqType.NT
) -> EntryValidationResults:
    def _set_validator_names(validation_results: EntryValidationResults) -> EntryValidationResults:
        return {f"correct_{region}{seq_type.value}_offsets": val for region, val in validation_results.items()}

    validation_results: EntryValidationResults = {region: None for region in REGIONS}

    if _getattr(rearrangement, "v_sequence_alignment", seq_type) is None:
        return _set_validator_names(validation_results)

    sequence = _getattr(rearrangement, "sequence", seq_type)
    fwr1_start = _getattr(rearrangement, "fwr1_start", seq_type)
    fwr1_end = _getattr(rearrangement, "fwr1_end", seq_type)
    if fwr1_start is not None and fwr1_end is not None and fwr1_end != -1:
        if fwr1_start == -1:
            fwr1_start = _getattr(rearrangement, "v_sequence_start", seq_type)
        validation_results["fwr1"] = sequence[fwr1_start - 1 : fwr1_end] == _getattr(rearrangement, "fwr1", seq_type)

    for region in REGIONS[1:-1]:
        # type: ignore
        region_start, region_end = _getattr(rearrangement, f"{region}_start", seq_type), _getattr(
            rearrangement, f"{region}_end", seq_type
        )
        if region_start is None or region_end is None or region_start == -1 or region_end == -1:
            continue
        validation_results[region] = sequence[region_start - 1 : region_end] == _getattr(
            rearrangement, region, seq_type
        )  # type: ignore

    if _getattr(rearrangement, "j_sequence_alignment", seq_type) is None:
        return _set_validator_names(validation_results)

    fwr4_start = _getattr(rearrangement, "fwr4_start", seq_type)
    fwr4_end = _getattr(rearrangement, "fwr4_end", seq_type)
    if fwr4_start != -1 and fwr4_start is not None and fwr4_end is not None:
        if fwr4_end == -1:
            fwr4_end = _getattr(rearrangement, "j_sequence_end", seq_type)
        validation_results["fwr4"] = sequence[fwr4_start - 1 : fwr4_end] == _getattr(rearrangement, "fwr4", seq_type)

    return _set_validator_names(validation_results)


def validate_correct_vj_in_frame(rearrangement: AirrRearrangementEntryNT) -> Optional[bool]:
    j_alignment_start, j_frame = rearrangement.j_alignment_start, rearrangement.j_frame
    v_alignment_start, v_frame = rearrangement.v_alignment_start, rearrangement.v_frame

    if (
        j_alignment_start is None
        or j_frame is None
        or v_alignment_start is None
        or v_frame is None
        or rearrangement.vj_in_frame is None
    ):
        return None

    vj_in_frame = ((j_alignment_start - 1 + j_frame) - (v_alignment_start - 1 + v_frame)) % 3 == 0

    return vj_in_frame == rearrangement.vj_in_frame


def validate_cdr3_in_junction(rearrangement: AirrRearrangementEntryNT) -> Optional[bool]:
    cdr3 = rearrangement.cdr3
    junction = rearrangement.junction
    cdr3_aa = rearrangement.cdr3_aa
    junction_aa = rearrangement.junction_aa
    if (not cdr3) or (not junction) or (not cdr3_aa) or (not junction_aa):
        return None
    return cdr3 in junction and cdr3_aa in junction_aa


def validate_cdr3_aa_in_junction_aa(rearrangement: AirrRearrangementEntryAA) -> Optional[bool]:
    cdr3_aa = rearrangement.cdr3_aa
    junction_aa = rearrangement.junction_aa
    if (not cdr3_aa) or (not junction_aa):
        return None
    return cdr3_aa in junction_aa


def validate_locus_as_in_v_gene(rearrangement: AirrRearrangementEntry_co) -> Optional[bool]:
    locus = rearrangement.locus
    v_call = rearrangement.v_call
    if not (locus) or (not v_call):
        return None
    return locus.upper() in v_call.upper()


def validate_gene_alignment(
    rearrangement: AirrRearrangementEntry_co, gene: Gene, seq_type: SeqType = SeqType.NT
) -> Optional[bool]:
    sequence = _getattr(rearrangement, "sequence", seq_type)
    sequence_start = _getattr(rearrangement, "sequence_start", seq_type, gene)
    sequence_end = _getattr(rearrangement, "sequence_end", seq_type, gene)
    sequence_alignment = _getattr(rearrangement, "sequence_alignment", seq_type, gene)

    if (not sequence_start) or (not sequence_end) or (not sequence_alignment):
        return None
    return sequence[sequence_start - 1 : sequence_end] == sequence_alignment


def validate_no_empty_regions_in_gene_alignment(
    rearrangement: AirrRearrangementEntry_co, gene: Gene, seq_type: SeqType = SeqType.NT
) -> EntryValidationResults:
    def _set_validator_names(validation_results: EntryValidationResults) -> EntryValidationResults:
        return {
            f"no_empty_{region}{seq_type.value}_in_{gene.value.rstrip('_')}": val
            for region, val in validation_results.items()
        }

    gene_full_regions = [
        region
        for region, start_end_gene in REGION_TO_GENE.items()
        if start_end_gene["start"] == start_end_gene["end"] == gene
    ]
    validation_results: EntryValidationResults = {region: None for region in gene_full_regions}
    gene_sequence_alignment = _getattr(rearrangement, "sequence_alignment", seq_type, gene)
    if gene_sequence_alignment is None:
        return _set_validator_names(validation_results)
    for region in gene_full_regions:
        if _getattr(rearrangement, region, seq_type) is not None:
            validation_results[region] = True
        else:
            validation_results[region] = False
    return _set_validator_names(validation_results)


def validate_no_empty_cdr3(rearrangement: AirrRearrangementEntry_co, seq_type: SeqType = SeqType.NT) -> Optional[bool]:
    if (
        _getattr(rearrangement, "v_sequence_alignment", seq_type) is None
        or _getattr(rearrangement, "j_sequence_alignment", seq_type) is None
    ):
        return None
    return _getattr(rearrangement, "cdr3", seq_type) is not None


def validate_cdr3_starts_inside_v_alignment(
    rearrangement: AirrRearrangementEntry_co, seq_type: SeqType = SeqType.NT
) -> Optional[bool]:
    cdr3_start = _getattr(rearrangement, "cdr3_start", seq_type)
    v_sequence_end = _getattr(rearrangement, "v_sequence_end", seq_type)
    if cdr3_start is None or v_sequence_end is None:
        return None
    if cdr3_start == -1 or cdr3_start > v_sequence_end:
        return False
    return True


def validate_conserved_residues_present(rearrangement: AirrRearrangementEntry_co) -> EntryValidationResults:
    imgt_conserved_residues = (
        IMGT_CONSERVED_RESIDUES_HEAVY if rearrangement.locus == "igh" else IMGT_CONSERVED_RESIDUES_LIGHT
    )
    numbering = rearrangement.scheme_residue_mapping
    validation_results: EntryValidationResults = {
        f"conserved_{aa}{pos}_present": None for pos, aa in imgt_conserved_residues.items()
    }
    if numbering is None:
        return validation_results

    return {
        f"conserved_{aa}{pos}_present": numbering.get(pos.split("_", maxsplit=1)[0]) == aa
        for pos, aa in imgt_conserved_residues.items()
    }


def validate_primary_sequence_in_sequence_alignment_aa(rearrangement: AirrRearrangementEntry_co) -> Optional[bool]:
    if rearrangement.sequence_alignment_aa is None or rearrangement.scheme_residue_mapping is None:
        return None
    return "".join(rearrangement.scheme_residue_mapping.values()) in rearrangement.sequence_alignment_aa


def validate_no_insertion_next_to_deletion_aa(rearrangement: AirrRearrangementEntry_co) -> Optional[bool]:
    if rearrangement.sequence_aa_scheme_cigar is None:
        return None
    alignment_str = unfold_cigar(rearrangement.sequence_aa_scheme_cigar)
    return "ID" not in alignment_str and "DI" not in alignment_str


def validate_insertions_in_correct_places(rearrangement: AirrRearrangementEntry_co) -> Optional[bool]:
    if rearrangement.scheme_residue_mapping is None:
        return None

    scheme = Scheme(rearrangement.numbering_scheme)
    chain_type = ChainType.from_locus(Locus(rearrangement.locus))

    allowed_cdr_insertions: set[int] = set(get_indels_positions(scheme, chain_type))
    if scheme == Scheme.IMGT:
        allowed_cdr_insertions = {*allowed_cdr_insertions, *[ins + 1 for ins in allowed_cdr_insertions]}

    allowed_framework_insertions = get_framework_insertion_positions(scheme, chain_type)
    for scheme_pos in rearrangement.scheme_residue_mapping:
        split = scheme_pos.split(".")
        if len(split) > 1:
            if (int(split[0]) not in allowed_cdr_insertions) and (scheme_pos not in allowed_framework_insertions):
                return False
    return True


VALIDATORS: dict[str, Callable[[AirrRearrangementEntryNT], Optional[bool]]] = {
    "regions_in_aligned_sequence": validate_regions_in_aligned_sequence,
    "regions_aa_in_aligned_sequence_aa": partial(validate_regions_in_aligned_sequence, seq_type=SeqType.AA),
    "translated_regions_in_aligned_sequence_aa": validate_translated_regions_in_aligned_sequence_aa,
    "correct_vj_in_frame": validate_correct_vj_in_frame,
    "cdr3_in_junction": validate_cdr3_in_junction,
    "locus_as_in_v_gene": validate_locus_as_in_v_gene,
    "v_gene_alignment": partial(validate_gene_alignment, gene=Gene.V),
    "j_gene_alignment": partial(validate_gene_alignment, gene=Gene.J),
    "c_gene_alignment": partial(validate_gene_alignment, gene=Gene.C),
    "no_negative_offsets_inside_v_alignment": partial(validate_no_negative_offsets_inside_gene_alignment, gene=Gene.V),
    "no_negative_offsets_inside_j_alignment": partial(validate_no_negative_offsets_inside_gene_alignment, gene=Gene.J),
    "consecutive_offsets": validate_consecutive_offsets,
    "no_empty_cdr3": validate_no_empty_cdr3,
    "primary_sequence_in_sequence_alignment_aa": validate_primary_sequence_in_sequence_alignment_aa,
    "no_insertion_next_to_deletion_aa": validate_no_insertion_next_to_deletion_aa,
    "insertions_in_correct_places": validate_insertions_in_correct_places,
}
MULTI_VALIDATORS: list[Callable[[AirrRearrangementEntryNT], EntryValidationResults]] = [
    validate_correct_region_offsets,
    partial(validate_no_empty_regions_in_gene_alignment, gene=Gene.V),
    partial(validate_no_empty_regions_in_gene_alignment, gene=Gene.J),
]


def validate_airr_entry(rearrangement: AirrRearrangementEntryNT, scheme: Scheme) -> dict[str, Optional[bool]]:
    entry_validation_results = {}
    for validator_name, validator in VALIDATORS.items():
        try:
            entry_validation_results[validator_name] = validator(rearrangement)
        except Exception:  # pylint: disable=broad-except
            pass

    for multi_validator in MULTI_VALIDATORS:
        try:
            partial_validation_result = multi_validator(rearrangement)
            entry_validation_results.update(partial_validation_result)
        except Exception:  # pylint: disable=broad-except
            pass

    try:
        match scheme:
            case Scheme.IMGT:
                partial_validation_result = validate_conserved_residues_present(rearrangement)
                entry_validation_results.update(partial_validation_result)
            case _:
                pass
    except Exception:  # pylint: disable=broad-except
        pass

    return entry_validation_results


VALIDATORS_AA: dict[str, Callable[[AirrRearrangementEntryAA], Optional[bool]]] = {
    "regions_aa_in_aligned_sequence_aa": partial(validate_regions_in_aligned_sequence, seq_type=SeqType.AA),
    "cdr3_aa_in_junction_aa": partial(validate_cdr3_in_junction, seq_type=SeqType.AA),
    "locus_as_in_v_gene": validate_locus_as_in_v_gene,
    "v_gene_alignment_aa": partial(validate_gene_alignment, gene=Gene.V, seq_type=SeqType.AA),
    "j_gene_alignment_aa": partial(validate_gene_alignment, gene=Gene.J, seq_type=SeqType.AA),
    "c_gene_alignment_aa": partial(validate_gene_alignment, gene=Gene.C, seq_type=SeqType.AA),
    "no_negative_offsets_inside_v_alignment_aa": partial(
        validate_no_negative_offsets_inside_gene_alignment, gene=Gene.V, seq_type=SeqType.AA
    ),
    "no_negative_offsets_inside_j_alignment_aa": partial(
        validate_no_negative_offsets_inside_gene_alignment, gene=Gene.J, seq_type=SeqType.AA
    ),
    "consecutive_offsets_aa": partial(validate_consecutive_offsets, seq_type=SeqType.AA),
    "no_empty_cdr3_aa": partial(validate_no_empty_cdr3, seq_type=SeqType.AA),
    "primary_sequence_in_sequence_alignment_aa": validate_primary_sequence_in_sequence_alignment_aa,
    "no_insertion_next_to_deletion_aa": validate_no_insertion_next_to_deletion_aa,
    "insertions_in_correct_places": validate_insertions_in_correct_places,
}
MULTI_VALIDATORS_AA: list[Callable[[AirrRearrangementEntryAA], EntryValidationResults]] = [
    partial(validate_correct_region_offsets, seq_type=SeqType.AA),
    partial(validate_no_empty_regions_in_gene_alignment, gene=Gene.V, seq_type=SeqType.AA),
    partial(validate_no_empty_regions_in_gene_alignment, gene=Gene.J, seq_type=SeqType.AA),
]


def validate_airr_entry_aa(rearrangement: AirrRearrangementEntryAA, scheme: Scheme) -> dict[str, Optional[bool]]:
    entry_validation_results = {}
    for validator_name, validator in VALIDATORS_AA.items():
        try:
            entry_validation_results[validator_name] = validator(rearrangement)  # type: ignore
        except Exception:  # pylint: disable=broad-except
            pass

    for multi_validator in MULTI_VALIDATORS_AA:
        try:
            partial_validation_result = multi_validator(rearrangement)  # type: ignore
            entry_validation_results.update(partial_validation_result)
        except Exception:  # pylint: disable=broad-except
            pass

    try:
        match scheme:
            case Scheme.IMGT:
                partial_validation_result = validate_conserved_residues_present(rearrangement)
                entry_validation_results.update(partial_validation_result)
            case _:
                pass
    except Exception:  # pylint: disable=broad-except
        pass

    return entry_validation_results
